"""ViPErLEED utilities for converting beam files to EXPBEAMS.csv."""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2024-03-27'
__license__ = 'GPLv3+'

from viperleed.utilities.beams.cli import BeamsUtilCLI

BeamsUtilCLI.run_as_script()
