"""Package beams of viperleed.utilities.

Collects utilities for converting beams in different formats to the
EXPBEAMS.csv that is used in ViPErLEED.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2024-03-27'
__license__ = 'GPLv3+'

EXPBEAMS_DEFAULT = 'EXPBEAMS.csv'
