"""ViPErLEED poscar utilities."""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-04'
__license__ = 'GPLv3+'

from viperleed.utilities.poscar.cli import PoscarUtilsCLI


PoscarUtilsCLI.run_as_script()
